import glob
import os
import sys
from typing import Literal

import numpy as np
import tensorflow as tf
import tf_levenberg_marquardt as lm
from natsort import natsorted

from cqpes.types import TrainConfig
from cqpes.utils.model import build_network

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

tf.config.set_visible_devices([], "GPU")
tf.keras.backend.set_floatx("float64")


def _isru(x):
    return x / tf.sqrt(tf.square(x) + 1.0)


tf.keras.utils.get_custom_objects().update({"isru": _isru})


_ACT_MAP = {
    "linear": 0,
    "tanh": 1,
    "sigmoid": 2,
    "isru": 3,
}


def _model2potfit(
    model: tf.keras.Model,
    output_dir: str,
    p_min: np.ndarray,
    p_max: np.ndarray,
    V_min: float,
    V_max: float,
) -> tuple[str, str]:
    weights_file = os.path.join(output_dir, "weights.txt")
    biases_file = os.path.join(output_dir, "biases.txt")

    # weights.txt
    with open(weights_file, "w") as f:
        print(
            f"{model.input_shape[1]:8d}"
            f"{(len(model.layers) - 1):8d}"
            f"{model.output_shape[1]:8d}",
            file=f,
        )

        for idx, layer in enumerate(model.layers):
            if idx == (len(model.layers) - 1):
                print("", file=f)

                break

            units = layer.get_config()["units"]

            print(f"{units:8d}", end="", file=f)

        act_name = model.layers[0].get_config()["activation"]
        act_id = _ACT_MAP.get(act_name)

        if act_id is None:
            raise ValueError(
                f"Activation '{act_name}' not supported for txt export."
            )

        print(f"{act_id:8d}{model.count_params():8d}", file=f)

        pdela = 0.5 * (
            np.concatenate((p_max[1:], [V_max]))
            - np.concatenate((p_min[1:], [V_min]))
        )

        pavga = 0.5 * (
            np.concatenate((p_max[1:], [V_max]))
            + np.concatenate((p_min[1:], [V_min]))
        )

        for x in pdela:
            print(f"{x:20.16f}\t", end="", file=f)

        print("", file=f)

        for x in pavga:
            print(f"{x:20.16f}\t", end="", file=f)

        print("", file=f)

        for layer in model.layers:
            W, _ = layer.get_weights()

            for w in W.T.flatten():
                print(f"{w:40.20f}", file=f)

    # bias.txt
    with open(biases_file, "w") as f:
        for layer in model.layers:
            _, b = layer.get_weights()

            for x in b:
                print(f"{x:40.20f}", file=f)

    return weights_file, biases_file


def run_export(
    workdir_path: str,
    export_type: Literal["h5", "potfit"],
) -> None:
    workdir = os.path.abspath(workdir_path)

    # 1. workdir
    config_path = os.path.join(workdir, "train.json")
    ckpt_dir = os.path.join(workdir, "ckpt")

    # 2. load train config
    train_config = TrainConfig.from_json(config_path)

    # 3. load scale factor
    try:
        p_min = np.load(os.path.join(workdir, "p_min.npy"))
        p_max = np.load(os.path.join(workdir, "p_max.npy"))
        V_min = np.load(os.path.join(workdir, "V_min.npy")).item()
        V_max = np.load(os.path.join(workdir, "V_max.npy")).item()
    except FileNotFoundError as e:
        print(
            f"\n[ERROR] Physical parameters missing in workdir: {e}",
            file=sys.stderr,
        )
        print("Did you forget to save .npy files during 'train'?")
        return

    # 4. auto detect checkpoint
    h5_files = natsorted(glob.glob(os.path.join(ckpt_dir, "*.weights.h5")))

    if not h5_files:
        print(
            f"\n[ERROR] No checkpoints found in {ckpt_dir}",
            file=sys.stderr,
        )
        print(
            "Did you start training?",
            file=sys.stderr,
        )
        return

    best_ckpt = h5_files[-1]

    # 5. build model
    input_dim = len(p_min) - 1
    model = build_network(train_config, input_dim=input_dim)
    model_wrapper = lm.model.ModelWrapper(model)
    model_wrapper.build(input_shape=(1, input_dim))
    model_wrapper.load_weights(best_ckpt)

    # 6. export
    export_dir = os.path.join(workdir, "export")
    os.makedirs(export_dir, exist_ok=True)

    if export_type == "h5":
        h5_path = os.path.join(export_dir, "model.h5")
        model_wrapper.model.save(h5_path)
        print(
            f"  [{'H5':^10}] Saved pure Keras model to: "
            f"{os.path.basename(h5_path)}"
        )
    elif export_type == "potfit":
        weights_path, biases_path = _model2potfit(
            model_wrapper.model, export_dir, p_min, p_max, V_min, V_max
        )
        print(
            f"  [{'POTFIT':^10}] Saved potfit model to "
            f"{os.path.basename(weights_path)} and "
            f"{os.path.basename(biases_path)}"
        )

    print(f"  [{'DONE':^10}] Results saved to {export_dir}")



from typing import Optional, TextIO, Type, Union

__all__ = [
    "print_header",
    "custom_warning",
]

_LOG_HEADER = r"""
                      _____ ____  _____  ______  _____                          
                     / ____/ __ \|  __ \|  ____|/ ____|                         
                    | |   | |  | | |__) | |__  | (___                           
                    | |   | |  | |  ___/|  __|  \___ \                          
                    | |___| |__| | |    | |____ ____) |                         
                     \_____\___\_\_|    |______|_____/                          
"""


def print_header(
    module_name: str,
    no_header: bool = False,
) -> None:
    if not no_header:
        print(f"{_LOG_HEADER:^80}")
        print(f"{'CQPES: ChongQing Potential Energy Surface':^80}")

    print()
    print(f"{module_name.upper():^80}")
    print("=" * 80)


def custom_warning(
    message: Union[str, Warning],
    category: Type[Warning],
    filename: str,
    lineno: int,
    file: Optional[TextIO] = None,
    line: Optional[str] = None,
) -> None:
    print(f"  [{'WARN':^10}] {message}")



import tensorflow as tf

from cqpes.types.train import TrainConfig


def build_network(
    config: TrainConfig,
    input_dim: int,
) -> tf.keras.Model:
    model = tf.keras.Sequential()
    model.add(tf.keras.Input(shape=(input_dim,)))

    for num_units in config.network.layers:
        model.add(
            tf.keras.layers.Dense(
                units=num_units,
                activation=config.network.activation,
            )
        )

    model.add(tf.keras.layers.Dense(units=1, activation="linear"))

    return model



import os
import sys
from typing import List, Literal, cast

import numpy as np
import tensorflow as tf
from ase import Atoms
from ase.io import read, write
from ase.units import Bohr, Hartree

from cqpes import CQPESPot

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

tf.config.set_visible_devices([], "GPU")
tf.keras.backend.set_floatx("float64")


def run_predict(
    workdir_path: str,
    xyz_path: str,
    output_path: str,
    return_au: bool = False,
    calc_forces: bool = True,
    force_mode: Literal["analytical", "numerical"] = "analytical",
) -> None:
    try:
        pot = CQPESPot(workdir_path, force_mode=force_mode)

        print(f"  [{'READ':^10}] Target: {os.path.basename(xyz_path)}")
        mol_list = cast(List[Atoms], read(xyz_path, index=":"))
        print(f"  [{'SAMPLES':^10}] Loaded {len(mol_list)} frames.")

        print(f"  [{'COMPUTE':^10}] Neural Network forward passing (Energy)...")

        energies = pot.get_energy(mol_list, return_au=return_au)
        if np.isscalar(energies):
            energies = [energies]

        if calc_forces:
            print(
                f"  [{'COMPUTE':^10}] Engine resolving forces ({force_mode})..."
            )

            forces_batch = pot.get_forces(mol_list)

            if forces_batch.ndim == 2:
                forces_batch = np.expand_dims(forces_batch, axis=0)

        if return_au:
            print(
                f"  [{'SCALE':^10}] Converting to Atomic Units (Bohr, Hartree)..."
            )
            unit_e, unit_c, unit_f = "Hartree", "Bohr", "Hartree/Bohr"

            # 1 eV/A = (Bohr / Hartree) a.u.
            if calc_forces:
                forces_batch *= Bohr / Hartree
        else:
            unit_e, unit_c, unit_f = "eV", "Angstrom", "eV/Angstrom"

        print(f"  [{'PACK':^10}] Injecting properties into ExtXYZ format...")
        for i, atoms in enumerate(mol_list):
            # to au
            if return_au:
                atoms.positions /= Bohr

                if atoms.cell.any():  # type: ignore
                    atoms.cell /= Bohr  # type: ignore

            atoms.info["energy"] = energies[i]
            atoms.info["energy_unit"] = unit_e
            atoms.info["length_unit"] = unit_c

            if calc_forces:
                atoms.arrays["forces"] = forces_batch[i]  # type: ignore
                atoms.info["force_unit"] = unit_f

        write(output_path, mol_list, format="extxyz")

        print(
            f"  [{'DONE':^10}] ExtXYZ trajectory saved to "
            f"{os.path.basename(output_path)}"
        )

    except Exception as e:
        print(f"\n[ERROR] Prediction failed: {e}", file=sys.stderr)
        sys.exit(1)



import glob
import os
import re

import numpy as np
import pandas as pd
import scienceplots  # noqa: F401
import tensorflow as tf
import tf_levenberg_marquardt as lm
from matplotlib import pyplot as plt
from natsort import natsorted
from sklearn.metrics import mean_absolute_error, mean_squared_error

from cqpes.types import CQPESData, TrainConfig
from cqpes.utils.model import build_network

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"

tf.config.set_visible_devices([], "GPU")
tf.keras.backend.set_floatx("float64")


def _isru(x):
    return x / tf.sqrt(tf.square(x) + 1.0)


tf.keras.utils.get_custom_objects().update({"isru": _isru})


def run_test(workdir_path: str):
    workdir = os.path.abspath(workdir_path)

    # load config
    train_config = TrainConfig.from_json(os.path.join(workdir, "train.json"))

    # load dataset
    dataset = CQPESData.from_dir(train_config.data)
    X_raw = dataset.X[:, 1:]
    V_true = dataset.V.reshape(-1, 1)

    params = {
        k: np.load(os.path.join(workdir, f"{k}.npy"))
        for k in ["p_min", "p_max", "V_min", "V_max"]
    }

    indices = {
        "Train": np.loadtxt(
            os.path.join(workdir, "train_idx.txt"), dtype=np.int32
        ),
        "Valid": np.loadtxt(
            os.path.join(workdir, "valid_idx.txt"), dtype=np.int32
        ),
        "Test": np.loadtxt(
            os.path.join(workdir, "test_idx.txt"), dtype=np.int32
        ),
    }

    # find checkpoint
    ckpt_dir = os.path.join(workdir, "ckpt")
    h5_files = natsorted(glob.glob(os.path.join(ckpt_dir, "*.h5")))

    if not h5_files:
        raise FileNotFoundError(f"[FATAL] No checkpoints found in {ckpt_dir}")

    best_ckpt_path = h5_files[-1]
    best_ckpt_name = os.path.basename(best_ckpt_path)

    match = re.search(r"epoch-[\w\d]+", best_ckpt_name)
    epoch_label = match.group(0) if match else "epoch-unknown"

    folder_name = os.path.basename(os.path.normpath(workdir))
    file_prefix = f"{folder_name}_{epoch_label}"

    print(f"  [{'WORKDIR':^10}] {workdir}")
    print(f"  [{'MODEL':^10}] Target: {best_ckpt_name}")

    # 4. build model
    model = build_network(train_config, input_dim=X_raw.shape[1])
    model_wrapper = lm.model.ModelWrapper(model)
    model_wrapper.build(input_shape=(1, X_raw.shape[1]))
    model_wrapper.load_weights(best_ckpt_path)

    # errors
    V_pred_norm = model_wrapper.predict(
        X_raw,
        batch_size=4096,
        verbose=0,
    )
    V_pred = CQPESData.unscale(V_pred_norm, params["V_min"], params["V_max"])
    errors_meV = (V_pred - V_true) * 1000.0

    # 5. summary
    _export_metrics(V_true, V_pred, indices, file_prefix)
    _plot_error_scatter(V_true, errors_meV, indices, file_prefix)
    _plot_error_dist(errors_meV, file_prefix)


def _export_metrics(V_true, V_pred, indices: dict, file_prefix: str) -> None:
    csv_filename = f"{file_prefix}_metrics.csv"
    stats = []

    eval_indices = {**indices, "Total": np.arange(len(V_true))}

    for name, idx in eval_indices.items():
        y_t, y_p = V_true[idx] * 1000.0, V_pred[idx] * 1000.0
        stats.append(
            {
                "Set": name,
                "MAE (meV)": mean_absolute_error(y_t, y_p),
                "RMSE (meV)": np.sqrt(mean_squared_error(y_t, y_p)),
                "MaxErr (meV)": np.abs(y_t - y_p).max(),
            }
        )

    df = pd.DataFrame(stats)
    df.to_csv(csv_filename, index=False)
    print(f"  [{'METRICS':^10}] Stats saved to: {csv_filename}")
    print("\n" + df.to_string(index=False) + "\n")


def _plot_error_scatter(
    V_true, errors_meV, indices: dict, file_prefix: str
) -> None:
    plot_filename = f"{file_prefix}_scatter.png"
    print(f"  [{'PLOT':^10}] Generating scatter plot...")

    colors = {"Train": "b", "Valid": "g", "Test": "r"}

    with plt.style.context(["science", "no-latex"]):
        fig, ax = plt.subplots(figsize=(8, 6), dpi=300)

        for name, idx in indices.items():
            ax.scatter(
                V_true[idx],
                errors_meV[idx],
                c=colors[name],
                alpha=0.5,
                label=name,
                s=1.0,
            )

        ax.axhline(0, color="#c0392b", linestyle="--", linewidth=1.5)
        ax.set_xlabel(r"$\mathrm{Ab \ Initio \ Energy \ (eV)}$")
        ax.set_ylabel(r"$\mathrm{Error \ (meV)}$")

        ax.legend(loc="upper right", frameon=True)

        plt.savefig(plot_filename, bbox_inches="tight")
        plt.close(fig)

        print(f"  [{'SAVE':^10}] Scatter plot saved as: {plot_filename}")


def _plot_error_dist(errors_meV, file_prefix: str) -> None:
    plot_filename = f"{file_prefix}_hist.png"
    print(f"  [{'PLOT':^10}] Generating histogram...")

    abs_err = np.abs(errors_meV)
    max_err = np.ceil(abs_err.max() * 2) / 2
    bin_width = 0.5 if max_err < 25 else max_err / 50
    edges = np.arange(0.0, max_err + bin_width, bin_width)

    with plt.style.context(["science", "no-latex"]):
        fig, ax = plt.subplots(figsize=(7, 5), dpi=300)

        weights = np.ones_like(abs_err) / len(abs_err)

        ax.hist(
            abs_err,
            bins=edges,  # type: ignore
            weights=weights,
            color="b",
            edgecolor="white",
            linewidth=0.8,
            rwidth=0.9,
        )

        ax.set_xlabel("Fitting Error (meV)", fontsize=12, fontweight="bold")
        ax.set_ylabel("Distribution", fontsize=12, fontweight="bold")

        plt.savefig(plot_filename, bbox_inches="tight")
        plt.close(fig)

        print(f"  [{'SAVE':^10}] Histogram saved as: {plot_filename}")



from .data import CQPESData
from .prepare import PrepareConfig, PrepareSummary
from .train import TrainConfig

__all__ = [
    "CQPESData",
    "PrepareConfig",
    "PrepareSummary",
    "TrainConfig",
]



import os
import shutil
from types import ModuleType
from typing import List, cast

import numpy as np
from ase import Atoms
from ase.io import read
from ase.units import Hartree
from scipy.spatial import distance

from cqpes.types import CQPESData, PrepareConfig, PrepareSummary
from cqpes.utils.msa import load_msa_so


def v_calc_p(
    xyz_list: np.ndarray,
    alpha: float,
    basis: ModuleType,
) -> np.ndarray:
    def _calc_morse(
        r: np.ndarray,
        alpha: float,
    ) -> np.ndarray:
        return np.exp(-1.0 * r / alpha)

    r_list = np.array([distance.pdist(xyz) for xyz in xyz_list])

    morse_list = np.apply_along_axis(
        func1d=_calc_morse,
        axis=1,
        arr=r_list,
        alpha=alpha,
    )

    mono_list = np.apply_along_axis(
        func1d=basis.evmono,
        axis=1,
        arr=morse_list,
    )

    poly_list = np.apply_along_axis(
        func1d=basis.evpoly,
        axis=1,
        arr=mono_list,
    )

    return poly_list


def v_calc_V(
    energy_list: np.ndarray,
    ref_energy: float,
) -> np.ndarray:
    V_list = (energy_list - ref_energy) * Hartree

    return V_list


def run_prepare(
    config: PrepareConfig,
    msa_path: str,
) -> PrepareSummary:
    # load msa so
    msa = load_msa_so(msa_path)
    basis = msa.basis

    # load xyz
    mol_list = cast(List[Atoms], read(config.xyz, index=":"))
    xyz_list = np.array([mol.get_positions() for mol in mol_list])

    # parse energy
    energy_list = np.loadtxt(config.energy)

    if len(xyz_list) != len(energy_list):
        raise ValueError(
            f"Dimension mismatch: xyz has {len(xyz_list)} frames, "
            f"but energy has {len(energy_list)} entries."
        )

    # p
    p_list = v_calc_p(
        xyz_list=xyz_list,
        alpha=config.alpha,
        basis=basis,
    )

    # ref energy
    if config.ref_energy is not None:
        ref_energy = config.ref_energy
    else:
        ref_energy = float(energy_list.min())

    V_list = v_calc_V(energy_list, ref_energy)

    # dataset
    cqpes_data = CQPESData(
        xyz=xyz_list,
        alpha=config.alpha,
        p=p_list,
        V=V_list,
        ref_energy=ref_energy,
    )

    output_path = cqpes_data.to_dir(config.output)

    assert cqpes_data.V_min is not None and cqpes_data.V_max is not None

    msa_filename = os.path.basename(msa_path)
    archived_msa = os.path.join(output_path, msa_filename)
    shutil.copy2(msa_path, archived_msa)

    return PrepareSummary(
        n_samples=cqpes_data.n_samples,
        n_atoms=xyz_list.shape[1],
        alpha=config.alpha,
        n_pip=p_list.shape[1],
        ref_energy=ref_energy,
        v_range=(cqpes_data.V_min, cqpes_data.V_max),
        output_dir=output_path,
    )



from typing import List, Literal

from ase.calculators.calculator import Calculator, all_changes

from cqpes.interface.potential import CQPESPot


class CQPESCalculator(Calculator):
    implemented_properties = ["energy", "forces"]

    default_parameters = {
        "force_mode": "analytical",
        "delta": 0.01,
    }

    def __init__(
        self,
        workdir: str,
        force_mode: Literal["analytical", "numerical"] = "analytical",
        delta: float = 0.01,  # Angstrom
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)

        if self.parameters is None:
            self.parameters = {}

        self.parameters.update(
            {
                "workdir": workdir,
                "force_mode": force_mode,
                "delta": delta,
            }
        )

        self.pot = CQPESPot(
            workdir=workdir,
            force_mode=force_mode,
        )

    def calculate(
        self,
        atoms=None,
        properties: List[str] = ["energy"],
        system_changes: List[str] = all_changes,
    ) -> None:
        super().calculate(atoms, properties, system_changes)

        self.results["energy"] = self.pot.get_energy(
            self.atoms,  # type: ignore
            return_au=False,
        )

        if "forces" in properties:
            self.results["forces"] = self.pot.get_forces(
                self.atoms,  # type: ignore
                force_mode=self.parameters["force_mode"],
                delta=self.parameters["delta"],
                return_au=False,
            )



import sys
from datetime import datetime

from ase import Atoms, units
from ase.constraints import FixCom
from ase.io import read, write
from ase.md.langevin import Langevin
from ase.md.logger import MDLogger
from ase.md.velocitydistribution import MaxwellBoltzmannDistribution
from ase.md.verlet import VelocityVerlet
from ase.optimize import BFGS
from ase.vibrations import Vibrations
from sella import Sella
from tqdm import tqdm

from cqpes.interface.ase import CQPESCalculator
from cqpes.utils.logger import print_header


def run_task(args) -> None:
    if (args.opt and args.md) or (args.freq and args.md):
        raise RuntimeError(
            "Only opt and freq tasks can be run sequentially.\n"
            "MD tasks are supposed to run individually."
        )

    # 1. read xyz
    data = read(args.xyz)
    atoms: Atoms = data[-1] if isinstance(data, list) else data

    # 2. setup calculator
    calc = CQPESCalculator(workdir=args.workdir, force_mode="analytical")
    atoms.calc = calc

    # 3. tasks
    output_basename = args.output.replace(".xyz", "")
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
    traj_file = None

    # (1) opt
    if args.opt:
        print_header("GEOMETRY OPTIMIZATION")

        if args.opt == "min":
            print(
                f"  [{'OPT':^10}] Mode: Minimization (BFGS) | fmax: {args.fmax}"
            )
            dyn = BFGS(atoms)
        else:
            print(f"  [{'TS':^10}] Mode: TS Search (Sella) | fmax: {args.fmax}")
            dyn = Sella(atoms)

        traj_file = f"{output_basename}_opt_{timestamp}.xyz"
        dyn.attach(lambda: write(traj_file, atoms, append=True))
        dyn.run(fmax=args.fmax, steps=args.steps)

    # (2) freq
    if args.freq:
        if args.opt:
            print_header("VIBRATIONAL ANALYSIS", no_header=True)
        else:
            print_header("VIBRATIONAL ANALYSIS")

        atoms.calc.parameters["force_mode"] = args.freq
        atoms.calc.pot.force_mode = args.freq

        vib_name = f"vib_{args.freq[:3]}"
        vib = Vibrations(atoms, name=vib_name, delta=args.freq_delta)

        print(
            f"  [{'FREQ':^10}] Mode: {args.freq.upper()} | "
            f"delta: {args.freq_delta} A"
        )

        vib.clean()
        vib.run()

        print(f"\n  [{'RESULT':^10}] Frequency Summary ({args.freq}):")

        vib.summary()

        atoms.calc.parameters["force_mode"] = "analytical"
        atoms.calc.pot.force_mode = "analytical"

    # (3) md
    if args.md:
        print_header("MOLECULAR DYNAMICS")

        MaxwellBoltzmannDistribution(atoms, temperature_K=args.temp)

        print(
            f"  [{'MD':^10}] Distribution: Maxwell-Boltzmann | "
            f"Temp: {args.temp} K"
        )

        dt_ase = args.dt * units.fs

        if args.md == "nve":
            # nve
            print(
                f"  [{'NVE':^10}] Ensemble: NVE (Velocity Verlet) | "
                f"dt: {args.dt} fs"
            )

            dyn_md = VelocityVerlet(
                atoms,
                timestep=dt_ase,
            )
        else:
            # nvt
            print(
                f"  [{'NVT':^10}] Ensemble: NVT (Langevin) | "
                f"Target: {args.temp} K"
            )

            atoms.set_constraint(FixCom())

            dyn_md = Langevin(
                atoms,
                timestep=dt_ase,
                temperature_K=args.temp,
                friction=0.01 / units.fs,
                fixcm=False,
            )

        traj_file = f"{output_basename}_md_{args.md}_{timestamp}.xyz"
        log_file = f"{output_basename}_md_{args.md}_{timestamp}.log"

        print(f"  [{'EXEC':^10}] Running {args.steps} MD steps...")

        # 1. traj
        dyn_md.attach(lambda: write(traj_file, atoms, append=True))

        # 2. logger
        logger = MDLogger(
            dyn_md,
            atoms,
            log_file,
            header=True,
            stress=False,
            peratom=False,
            mode="w",
        )

        dyn_md.attach(logger, interval=1)

        # 2. pbar
        pbar = tqdm(
            total=args.steps,
            desc=f"  [{'PROGRESS':^10}]",
            ncols=85,
            unit="step",
            file=sys.stdout,
            bar_format=(
                "{desc} {percentage:3.0f}% |{bar}| "
                "{n_fmt}/{total_fmt}[{elapsed}<{remaining}]"
            ),
        )

        md_state = {"is_step_zero": True}

        def update_pbar():
            if md_state["is_step_zero"]:
                md_state["is_step_zero"] = False
            else:
                pbar.update(1)

        dyn_md.attach(update_pbar)

        dyn_md.run(args.steps)

        pbar.close()

    if traj_file:
        print(f"  [{'SAVED':^10}] Trajectory saved to: {traj_file}")

        if args.md:
            print(f"  [{'SAVED':^10}] MD log saved to: {log_file}")

    print(f"  [{'DONE':^10}] Process completed")



import json
import os
import warnings
from dataclasses import dataclass, field, fields
from typing import Optional, Tuple


@dataclass
class PrepareConfig:
    # path
    xyz: str = field(metadata={"must_exist": True})
    energy: str = field(metadata={"must_exist": True})
    output: str = field(metadata={"must_not_exist": True})

    # phys
    alpha: float
    ref_energy: Optional[float] = None

    def __post_init__(
        self,
    ) -> None:
        for f in fields(self):
            val = getattr(self, f.name)

            if val is None:
                continue

            # to abs path
            if ("must_exist" in f.metadata) or ("must_not_exist" in f.metadata):
                val = os.path.abspath(val)
                setattr(self, f.name, val)

            # check xyz and energy
            if f.metadata.get("must_exist") and not os.path.exists(val):
                raise FileNotFoundError(
                    f"[{f.name}] Input file '{val}' not found."
                )

            # check output
            if f.metadata.get("must_not_exist") and os.path.exists(val):
                raise FileExistsError(
                    f"[{f.name}] Output directory '{val}' already exists."
                )

        # check if reference energy exists
        if self.ref_energy is None:
            warnings.warn("Use minimum energy as reference energy.")

        # check if alpha > 0
        if self.alpha <= 0:
            raise ValueError(f"alpha must be greater than 0, got {self.alpha}.")

    @classmethod
    def from_json(
        cls,
        config_json: str,
    ) -> "PrepareConfig":
        if not os.path.exists(config_json):
            raise FileNotFoundError(f"Config file '{config_json}' not found.")

        with open(config_json, "r") as f:
            config_dict = json.load(f)

        valid_keys = {f.name for f in fields(cls)}
        filtered_kwargs = {
            k: v for k, v in config_dict.items() if k in valid_keys
        }

        return cls(**filtered_kwargs)


@dataclass(frozen=True)
class PrepareSummary:
    n_samples: int
    n_atoms: int
    alpha: float
    n_pip: int
    ref_energy: float
    v_range: Tuple[float, float]
    output_dir: str

    def log(self) -> None:
        WIDTH = 80

        print(f"  {'Samples:':<16} {self.n_samples}")
        print(f"  {'Atoms:':<16} {self.n_atoms}")
        print(f"  {'Morse Alpha:':<16} {self.alpha:<16.4f}")
        print(f"  {'PIP Dimension:':<16} {self.n_pip}")

        print(f"  {'-' * (WIDTH - 4)}")

        print(f"  {'Ref Energy:':<16} {self.ref_energy:<16.8f} Hartree")
        print(f"  {'V_min (eV):':<16} {self.v_range[0]:<16.4f}")
        print(f"  {'V_max (eV):':<16} {self.v_range[1]:<16.4f}")

        print(f"  {'-' * (WIDTH - 4)}")

        print(f"  {'Output Dir:':<16}")
        print(f"    {self.output_dir}")

        print("=" * WIDTH)



import datetime
import glob
import importlib.util
import inspect
import json
import os
import shutil
from dataclasses import asdict
from typing import Callable

import numpy as np
import tensorflow as tf
import tf_levenberg_marquardt as lm
from sklearn.model_selection import train_test_split
from tensorflow.keras.callbacks import ModelCheckpoint  # type: ignore
from tensorflow.keras.callbacks import TensorBoard

from cqpes.types import CQPESData, TrainConfig
from cqpes.utils.model import build_network

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "2"
tf.keras.backend.set_floatx("float64")


def _isru(x):
    return x / tf.sqrt(tf.square(x) + 1.0)


tf.keras.utils.get_custom_objects().update({"isru": _isru})


def _load_weighting_func() -> Callable:
    def default_weighting(
        v: float,
    ) -> float:
        return 1.0

    weight_path = os.path.abspath("weighting.py")

    if not os.path.exists(weight_path):
        return default_weighting

    spec = importlib.util.spec_from_file_location(
        "weighting_module", weight_path
    )

    if spec and spec.loader:
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)

        if hasattr(module, "weighting"):
            return module.weighting
        else:
            raise AttributeError(
                f"[FATAL] Found custom weighting file at '{weight_path}', "
                "but it does NOT define the required 'weighting' function.\n"
                "Please define 'def weighting(v):' "
                "or remove the file to use defaults."
            )

    raise ImportError(f"[FATAL] Failed to load module from '{weight_path}'.")


def run_train(
    config: TrainConfig,
) -> None:
    WIDTH = 80

    # 1. workdir
    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")
    run_workdir = f"{config.workdir}_{timestamp}"
    os.makedirs(run_workdir, exist_ok=True)

    print(f"  [{'WORKDIR':^10}] {run_workdir}")

    config_snapshot = os.path.join(run_workdir, "train.json")

    with open(config_snapshot, "w") as f:
        json.dump(asdict(config), f, indent=4)

    msa_files = glob.glob(os.path.join(config.data, "*.so"))

    for msa_file in msa_files:
        shutil.copy2(
            src=msa_file,
            dst=os.path.join(run_workdir, os.path.basename(msa_file)),
        )

    weight_src = os.path.abspath("weighting.py")

    if os.path.exists(weight_src):
        shutil.copy(weight_src, os.path.join(run_workdir, "weighting.py"))

    # 2. load data & split
    print(f"  [{'LOAD':^10}] Loading CQPES dataset from {config.data}...")

    dataset = CQPESData.from_dir(config.data)

    X = dataset.X[:, 1:]
    y = dataset.y
    V = dataset.V

    artifacts = [
        "alpha.npy",
        "ref_energy.npy",
        "p_min.npy",
        "p_max.npy",
        "V_min.npy",
        "V_max.npy",
    ]

    for item in artifacts:
        src = os.path.join(config.data, item)
        if os.path.exists(src):
            shutil.copy2(src, run_workdir)

    index = np.arange(len(X)).astype(np.int32)
    r_train, r_val, r_test = config.split

    train_idx, valid_idx = train_test_split(index, test_size=(1 - r_train))
    valid_idx, test_idx = train_test_split(
        valid_idx, test_size=(r_test / (1 - r_train))
    )

    np.savetxt(
        os.path.join(run_workdir, "train_idx.txt"),
        train_idx,
        fmt="%d",
    )

    np.savetxt(
        os.path.join(run_workdir, "valid_idx.txt"),
        valid_idx,
        fmt="%d",
    )

    np.savetxt(
        os.path.join(run_workdir, "test_idx.txt"),
        test_idx,
        fmt="%d",
    )

    print(
        f"  [{'SPLIT':^10}] Train: {len(train_idx)} | "
        f"Valid: {len(valid_idx)} | "
        f"Test: {len(test_idx)}"
    )

    # 3. weighting
    weighting_func = _load_weighting_func()

    print(f"  [{'WEIGHT':^10}] Active weighting function:")
    print("\n")

    for line in inspect.getsource(weighting_func).strip().split("\n"):
        print(f"             {line}")

    print("\n")

    # for better performance
    weights = np.fromiter(
        (weighting_func(v) for v in V),
        dtype=np.float64,
        count=len(V),
    )

    # 4. build nn
    print(f"  [{'NETWORK':^10}] Constructing PIP-NN with LM Optimizer...")

    model = build_network(config, input_dim=X.shape[1])

    # 5. lm optimizer
    model_wrapper = lm.model.ModelWrapper(model)
    model_wrapper.compile(
        optimizer=tf.keras.optimizers.SGD(learning_rate=config.fit.lr),
        loss=lm.loss.MeanSquaredError(),
        damping_algorithm=lm.damping.DampingAlgorithm(
            adaptive_scaling=config.lm.adaptive_scaling,
            fletcher=config.lm.fletcher,
        ),
        solve_method=config.lm.solve_method,
        jacobian_max_num_rows=config.lm.jacobian_max_num_rows,
        metrics=[tf.keras.metrics.MeanSquaredError(name="mse")],
        weighted_metrics=[tf.keras.metrics.MeanSquaredError(name="wmse")],
    )

    # 6. config callbacks
    ckpt_dir = os.path.join(run_workdir, "ckpt")
    log_dir = os.path.join(run_workdir, "log")
    os.makedirs(ckpt_dir, exist_ok=True)
    os.makedirs(log_dir, exist_ok=True)

    ckpt = ModelCheckpoint(
        filepath=os.path.join(
            ckpt_dir,
            "model_epoch_{epoch:04d}_val_mse_{val_wmse:.5e}.weights.h5",
        ),
        monitor="val_wmse",
        save_best_only=True,
        save_weights_only=True,
    )

    tensorboard = TensorBoard(log_dir)

    # 7. start
    batch_size = (
        len(train_idx) if config.fit.batch_size == -1 else config.fit.batch_size
    )

    print(
        f"  [{'TRAIN':^10}] Starting LM training loop | "
        f"Epochs: {config.fit.epoch} | "
        f"Batch Size: {batch_size}"
    )

    print("-" * WIDTH)

    model_wrapper.fit(
        X[train_idx],
        y[train_idx],
        batch_size=batch_size,
        epochs=config.fit.epoch,
        sample_weight=weights[train_idx],
        validation_data=(X[valid_idx], y[valid_idx], weights[valid_idx]),
        callbacks=[tensorboard, ckpt],
        verbose=2,  # type: ignore
    )



import glob
import os
from typing import List, Literal, Optional, Union

import numpy as np
import tensorflow as tf
from ase import Atoms
from ase.units import Bohr, Hartree
from scipy.spatial.distance import pdist

from cqpes.pipeline.prepare import v_calc_p
from cqpes.types.data import CQPESData
from cqpes.types.train import TrainConfig
from cqpes.utils.model import build_network
from cqpes.utils.msa import load_msa_so

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

tf.get_logger().setLevel("ERROR")

tf.config.set_visible_devices([], "GPU")
tf.keras.backend.set_floatx("float64")


def _isru(x):
    return x / tf.sqrt(tf.square(x) + 1.0)


tf.keras.utils.get_custom_objects().update({"isru": _isru})


class CQPESPot:
    def __init__(
        self,
        workdir: str,
        force_mode: Literal["analytical", "numerical"] = "analytical",
    ) -> None:
        self.workdir = os.path.abspath(workdir)
        self.force_mode = force_mode

        # load phys
        self._load_physics()

        # load pip in pip-nn
        self._mount_msa()

        # build nn in pip-nn
        self._build_net()

    def _load_physics(
        self,
    ) -> None:
        config_path = os.path.join(self.workdir, "train.json")
        if not os.path.exists(config_path):
            raise FileNotFoundError(
                f"[FATAL] train.json not found in {self.workdir}"
            )

        self.config = TrainConfig.from_json(config_path)

        try:
            self.p_min = np.load(os.path.join(self.workdir, "p_min.npy"))[1:]
            self.p_max = np.load(os.path.join(self.workdir, "p_max.npy"))[1:]

            self.V_min = np.load(os.path.join(self.workdir, "V_min.npy")).item()
            self.V_max = np.load(os.path.join(self.workdir, "V_max.npy")).item()

            self.alpha = np.load(os.path.join(self.workdir, "alpha.npy")).item()

            self.ref_energy = np.load(
                os.path.join(self.workdir, "ref_energy.npy")
            ).item()

        except Exception as e:
            raise RuntimeError(
                f"[ERROR] Failed to load physical artifacts: {e}"
            )

    def _mount_msa(self) -> None:
        so_files = glob.glob(os.path.join(self.workdir, "*.so"))

        if not so_files:
            raise FileNotFoundError(
                f"[FATAL] No MSA engine (.so) found in {self.workdir}"
            )

        msa_path = so_files[0]
        self.msa = load_msa_so(msa_path)

    def _build_net(self) -> None:
        export_model_path = os.path.join(self.workdir, "export", "model.h5")

        if not os.path.exists(export_model_path):
            raise FileNotFoundError(
                f"\n  [{'FATAL':^10}] Production model missing: "
                f"{os.path.basename(export_model_path)}\n"
                f"  [{'ACTION':^10}] Run: cqpes export -t h5 {self.workdir}"
            )

        input_dim = len(self.p_min)
        model = build_network(self.config, input_dim=input_dim)
        model.load_weights(export_model_path)

        self.net = model

    def _standardize_coordinates(
        self,
        xyz: Union[np.ndarray, List[Atoms], Atoms],
    ) -> np.ndarray:
        if isinstance(xyz, Atoms):
            xyz_input = xyz.get_positions()
        elif (
            isinstance(xyz, list) and len(xyz) > 0 and isinstance(xyz[0], Atoms)
        ):
            xyz_input = np.array([mol.get_positions() for mol in xyz])
        else:
            xyz_input = xyz

        arr = np.asarray(xyz_input, dtype=np.float64)

        if arr.ndim == 2:
            return arr[np.newaxis, :, :]
        elif arr.ndim == 3:
            return arr
        else:
            raise ValueError(
                f"[FATAL] Invalid coordinate shape {arr.shape}. "
                f"Expected (N, 3) or (M, N, 3)."
            )

    def get_energy(
        self,
        xyz: Union[np.ndarray, List[Atoms]],
        return_au: bool = False,
    ) -> np.ndarray:
        # (N_configs, N_atoms, 3)
        xyz_arr = self._standardize_coordinates(xyz)

        # 1. calculate p
        p_raw = v_calc_p(
            xyz_list=xyz_arr,
            alpha=self.alpha,
            basis=self.msa.basis,
        )

        p_feat = p_raw[:, 1:]

        # 2. p -> X in [-1, 1]
        x_scaled = CQPESData.rescale(p_feat, self.p_min, self.p_max)

        # 3. X -> y
        y_scaled = self.net.predict(
            x_scaled,
            batch_size=4096,
            verbose=0,
        )

        # 4. y in [-1, 1] -> V
        V_eV = CQPESData.unscale(y_scaled, self.V_min, self.V_max)

        results = V_eV.flatten()

        # 5. to Hartree
        if return_au:
            results = (results / Hartree) + self.ref_energy

        if results.size == 1 and not isinstance(xyz, list):
            return results.item()

        return results

    def get_forces(
        self,
        xyz: Union[np.ndarray, List[Atoms], Atoms],
        return_au: bool = False,
        force_mode: Optional[str] = None,
        **kwargs,
    ) -> np.ndarray:
        target_mode = (force_mode or self.force_mode).lower()

        if target_mode == "analytical":
            forces = self.get_forces_analytical(xyz)
        else:
            forces = self.get_forces_numerical(xyz, **kwargs)

        if return_au:
            forces *= Bohr / Hartree

        return forces

    def get_forces_numerical(
        self,
        xyz: Union[np.ndarray, List[Atoms], Atoms],
        delta: float = 0.01,
    ) -> np.ndarray:
        # (N_configs, N_atoms, 3)
        xyz_arr = self._standardize_coordinates(xyz)
        n_configs, n_atoms = xyz_arr.shape[0], xyz_arr.shape[1]

        n_displacements = 2 * 3 * n_atoms
        total_evals = n_configs * n_displacements
        displaced_batch = np.zeros((total_evals, n_atoms, 3), dtype=np.float64)

        idx = 0

        for c in range(n_configs):
            pos_ref = xyz_arr[c]
            for i in range(n_atoms):
                for j in range(3):
                    for d in [delta, -delta]:
                        displaced_batch[idx] = pos_ref.copy()
                        displaced_batch[idx, i, j] += d
                        idx += 1

        all_energies = self.get_energy(displaced_batch)

        # (N_configs, N_atoms, 3_cartesian, 2_directions)
        all_energies = all_energies.reshape(n_configs, n_atoms, 3, 2)

        # e_plus is index 0, e_minus is index 1 in the last dimension
        results_forces = -(all_energies[..., 0] - all_energies[..., 1]) / (
            2 * delta
        )

        if n_configs == 1 and not isinstance(xyz, list):
            return results_forces[0]

        return results_forces

    def get_forces_analytical(
        self,
        xyz: Union[np.ndarray, List[Atoms], Atoms],
    ) -> np.ndarray:
        # (N_configs, N_atoms, 3)
        xyz_arr = self._standardize_coordinates(xyz)
        n_configs = xyz_arr.shape[0]
        n_atoms = xyz_arr.shape[1]
        n_cart = 3 * n_atoms

        # 1. calculate p
        p_raw = v_calc_p(
            xyz_list=xyz_arr,
            alpha=self.alpha,
            basis=self.msa.basis,
        )

        p_feat = p_raw[:, 1:]

        # 2. prepare tensor and track gradients for the entire batch
        X_numpy = CQPESData.rescale(p_feat, self.p_min, self.p_max)
        X_tensor = tf.convert_to_tensor(X_numpy, dtype=tf.float64)

        with tf.GradientTape() as tape:
            tape.watch(X_tensor)
            y = self.net(X_tensor)

        grad_raw = tape.gradient(y, X_tensor)

        if grad_raw is None:
            raise RuntimeError(
                "[FATAL] Gradient disconnected. "
                "Ensure network topology is differentiable."
            )

        V_scale = (self.V_max - self.V_min) / 2.0
        p_scale = 2.0 / (self.p_max - self.p_min)

        # (N_configs, N_features)
        dV_dp_batch = tf.convert_to_tensor(grad_raw).numpy() * V_scale * p_scale

        results_forces = np.zeros((n_configs, n_atoms, 3), dtype=np.float64)

        for i in range(n_configs):
            pos = xyz_arr[i]
            dV_dp = dV_dp_batch[i]

            # 1. xyz -> dist vec
            r_vec = pdist(pos)
            n_bonds = len(r_vec)
            msa_drdx = np.zeros((n_cart, n_bonds), dtype=np.float64, order="F")

            k = 0

            for row in range(n_atoms - 1):
                for col in range(row + 1, n_atoms):
                    r = r_vec[k]
                    if r > 1e-12:
                        unit_vec = (pos[row] - pos[col]) / r
                        msa_drdx[3 * row : 3 * row + 3, k] = unit_vec
                        msa_drdx[3 * col : 3 * col + 3, k] = -unit_vec

                    k += 1

            # 2. dist vec -> morse -> mono -> poly
            x_morse = np.exp(-r_vec / self.alpha).ravel().astype(np.float64)
            msa_mono = self.msa.basis.evmono(x_morse)
            msa_poly = self.msa.basis.evpoly(msa_mono)

            # 3. dbemsav
            forces_flat = np.zeros(n_cart)

            for j in range(n_cart):
                # fortran 1-based indexing
                dp_dr_j = self.msa.gradient.dbemsav(
                    msa_drdx,
                    msa_mono,
                    msa_poly,
                    j + 1,
                )

                forces_flat[j] = -np.dot(dV_dp, dp_dr_j[1:])

            results_forces[i] = forces_flat.reshape(n_atoms, 3)

        if n_configs == 1 and not isinstance(xyz, list):
            return results_forces[0]

        return results_forces

    def check_forces(
        self,
        xyz: Union[np.ndarray, List[Atoms], Atoms],
        delta: float = 0.01,
        atol: float = 1.0e-05,
        rtol: float = 1.0e-03,
        verbose: bool = True,
    ) -> bool:
        f_ana = self.get_forces(xyz, force_mode="analytical")
        f_num = self.get_forces(xyz, force_mode="numerical", delta=delta)

        is_consistent = np.allclose(f_ana, f_num, atol=atol, rtol=rtol)

        if verbose:
            max_diff = np.max(np.abs(f_ana - f_num))
            status = "PASSED" if is_consistent else "FAILED"
            print(
                f"[CHECK_FORCES] Status: {status}\n"
                f"               MaxDiff: {max_diff:.4e}\n"
                f"               delta: {delta}"
            )

        return is_consistent



import json
import os
from dataclasses import dataclass, field, fields
from typing import Any, Dict, List, Literal, Type, TypeVar

import numpy as np

T = TypeVar("T")


def _filter_kwargs(
    cls: Type[T],
    kwargs: Dict[str, Any],
) -> Dict[str, Any]:
    valid_keys = {f.name for f in fields(cls)}  # type: ignore
    return {k: v for k, v in kwargs.items() if k in valid_keys}


@dataclass
class NetworkConfig:
    layers: List[int] = field(default_factory=lambda: [20, 20])
    activation: Literal["tanh", "isru", "sigmoid"] = "tanh"


@dataclass
class FitConfig:
    lr: float = 0.001
    epoch: int = 1000
    batch_size: int = -1  # full batch


@dataclass
class LMConfig:
    adaptive_scaling: bool = True
    fletcher: bool = False
    solve_method: Literal["solve", "lstsq"] = "solve"
    jacobian_max_num_rows: int = 500


@dataclass
class TrainConfig:
    data: str = field(metadata={"must_exist": True})
    workdir: str = field(metadata={"must_not_exist": False})
    network: NetworkConfig = field(default_factory=NetworkConfig)
    fit: FitConfig = field(default_factory=FitConfig)
    lm: LMConfig = field(default_factory=LMConfig)
    split: List[float] = field(default_factory=lambda: [0.9, 0.05, 0.05])

    def __post_init__(self) -> None:
        for f in fields(self):
            val = getattr(self, f.name)

            if not isinstance(val, str):
                continue

            # to abs path
            if ("must_exist" in f.metadata) or ("must_not_exist" in f.metadata):
                abs_val = os.path.abspath(val)
                setattr(self, f.name, abs_val)

                # check data
                if f.metadata.get("must_exist") and not os.path.exists(abs_val):
                    raise FileNotFoundError(
                        f"[{f.name}] Data directory '{abs_val}' not found. "
                        f"Please run 'cqpes prepare' first."
                    )

                # check workdir
                if f.metadata.get("must_not_exist") and os.path.exists(abs_val):
                    raise FileExistsError(
                        f"[{f.name}] Work directory '{abs_val}' already exists."
                    )

        if len(self.split) != 3:
            raise ValueError(
                f"split must contain exactly 3 elements "
                f"(Train, Valid, Test), got {len(self.split)}"
            )

        if not np.allclose(sum(self.split), 1.0):
            raise ValueError(
                f"split fractions must sum to 1.0, got {sum(self.split)}"
            )

    @classmethod
    def from_json(
        cls,
        config_json: str,
    ) -> "TrainConfig":
        if not os.path.exists(config_json):
            raise FileNotFoundError(f"Config file not found: {config_json}")

        with open(config_json, "r") as f:
            d = json.load(f)

        return cls(
            data=d["data"],
            workdir=d["workdir"],
            network=NetworkConfig(
                **_filter_kwargs(NetworkConfig, d.get("network", {}))
            ),
            fit=FitConfig(**_filter_kwargs(FitConfig, d.get("fit", {}))),
            lm=LMConfig(**_filter_kwargs(LMConfig, d.get("lm", {}))),
            split=d.get("split", [0.9, 0.05, 0.05]),
        )



import argparse
import os
import sys
import warnings
from importlib.metadata import PackageNotFoundError, version
from typing import Literal

os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"

try:
    import absl.logging

    absl.logging.set_verbosity(absl.logging.ERROR)

    import tensorflow as tf

    tf.get_logger().setLevel("ERROR")
except ImportError:
    pass

import argcomplete

from cqpes.utils.logger import custom_warning, print_header

warnings.showwarning = custom_warning

warnings.filterwarnings("ignore", message=".*HDF5 file.*")
warnings.filterwarnings("ignore", message=".*native Keras format.*")


def get_version() -> str:
    try:
        return version("cqpes")
    except PackageNotFoundError:
        return "unknown-dev"


def fuzzy_choice(choices):
    def wrapper(value):
        matches = [c for c in choices if c.startswith(value.lower())]
        
        if len(matches) == 1:
            return matches[0]

        elif len(matches) > 1:
            import argparse
            raise argparse.ArgumentTypeError(
                f"ambiguous value '{value}': could match {', '.join(matches)}"
            )
        else:
            import argparse
            raise argparse.ArgumentTypeError(
                f"invalid choice '{value}' (choose from {', '.join(choices)})"
            )

    return wrapper


def main() -> None:
    current_version = get_version()

    parser = argparse.ArgumentParser(
        prog="cqpes",
        description="CQPES: GPU-Aided Potential Energy Surface Development Toolkit",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument(
        "-v",
        "--version",
        action="version",
        version=f"%(prog)s {current_version}",
    )

    subparsers = parser.add_subparsers(
        dest="command",
        required=True,
        help="Sub-commands",
    )

    # 1. prepare
    prepare_parser = subparsers.add_parser(
        "prepare",
        help="Prepare dataset from raw files",
    )

    prepare_parser.add_argument(
        "config",
        help="Path to prepare.json",
    )

    prepare_parser.add_argument(
        "--msa",
        required=True,
        help="Path to compiled MSA .so module",
    )

    # 2. train
    train_parser = subparsers.add_parser(
        "train",
        help="Train PIP-NN model via Keras & TensorFlow",
    )

    train_parser.add_argument(
        "config",
        help="Path to train.json",
    )

    # 3. test
    test_parser = subparsers.add_parser(
        "test",
        help="Evaluate model performance and plot errors",
    )

    test_parser.add_argument(
        "workdir",
        help="Path to the workdir",
    )

    # 4. export
    export_parser = subparsers.add_parser(
        "export",
        help="Export trained model for dynamics interfaces",
    )

    export_parser.add_argument(
        "workdir",
        help="Path to the workdir",
    )

    export_parser.add_argument(
        "-t",
        "--type",
        choices=["h5", "potfit"],
        default="h5",
        help="Export format",
    )

    # 5. predict
    predict_parser = subparsers.add_parser(
        "predict",
        help="Predict energies for a given XYZ trajectory",
    )

    predict_parser.add_argument(
        "workdir",
        help="Path to the experiment workdir",
    )

    predict_parser.add_argument(
        "xyz",
        help="Path to the target .xyz file",
    )

    predict_parser.add_argument(
        "--no-forces",
        action="store_false",
        default=True,
        dest="calc_forces",
        help="Disable force calculation (Energy only)",
    )

    predict_parser.add_argument(
        "--force-mode",
        default="analytical",
        type=fuzzy_choice(["analytical", "numerical"]),
        help="Force calculation mode (default: analytical)",
    )

    predict_parser.add_argument(
        "-o",
        "--output",
        default="predicted.xyz",
        help="Output file path (ExtXYZ format)",
    )

    predict_parser.add_argument(
        "--au",
        action="store_true",
        default=False,
        dest="return_au",
        help=(
            "Output in Atomic Units "
            "(Energy: Hartree, Coords: Bohr, Forces: Hartree/Bohr)"
        ),
    )

    # 6. run tasks
    run_parser = subparsers.add_parser(
        "run",
        help="Run tasks (Opt, TS, Freq, MD) using trained PES via ASE",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    run_parser.add_argument(
        "workdir",
        help="Path to the model directory",
    )

    run_parser.add_argument(
        "xyz",
        help="Path to the initial .xyz file",
    )

    task_group = run_parser.add_argument_group("task selection")
    ctrl_group = run_parser.add_argument_group("control parameters")

    # (1) opt
    task_group.add_argument(
        "--opt",
        nargs="?",
        const="min",
        choices=["min", "ts"],
        help="Optimization mode: 'min' (minimization) or 'ts' (transition state)",
    )

    ctrl_group.add_argument(
        "--fmax",
        type=float,
        default=1.0e-05,
        help="Force threshold (eV/Ang) for Opt & TS",
    )

    # (2) freq
    task_group.add_argument(
        "--freq",
        nargs="?",
        const="analytical",
        type=fuzzy_choice(["analytical", "numerical"]),
        help="Frequency analysis: 'analytical' or 'numerical'",
    )

    ctrl_group.add_argument(
        "--freq-delta",
        type=float,
        default=0.01,
        help=(
            "Displacement (Ang) for frequency analysis.\n"
            "Gaussian-recommended value. (https://gaussian.com/freq)"
        ),
    )

    # (3) md
    task_group.add_argument(
        "--md",
        nargs="?",
        const="nve",
        choices=["nve", "nvt"],
        help="Molecular Dynamics mode",
    )

    ctrl_group.add_argument(
        "--temp",
        type=float,
        default=300.0,
        help="Temperature in K for MD",
    )

    ctrl_group.add_argument(
        "--dt",
        type=float,
        default=1.0,
        help="Time step in fs for MD",
    )

    # general
    ctrl_group.add_argument(
        "--steps",
        type=int,
        default=1000,
        help="Max steps for Opt, TS, or MD",
    )

    ctrl_group.add_argument(
        "-o",
        "--output",
        default="run_out.xyz",
        help="Output trajectory path: default is run_out.xyz",
    )

    argcomplete.autocomplete(parser)

    args = parser.parse_args()

    if args.command == "prepare":
        _run_prepare(
            args.config,
            msa_path=args.msa,
        )
    elif args.command == "train":
        _run_train(args.config)
    elif args.command == "test":
        _run_test(args.workdir)
    elif args.command == "export":
        _run_export(
            args.workdir,
            export_type=args.type,
        )
    elif args.command == "predict":
        _run_predict(
            args.workdir,
            args.xyz,
            output_path=args.output,
            return_au=args.return_au,
            calc_forces=args.calc_forces,
            force_mode=args.force_mode,
        )
    elif args.command == "run":
        _run_tasks(args)


def _run_prepare(
    config_path: str,
    msa_path: str,
) -> None:
    from cqpes.pipeline.prepare import run_prepare
    from cqpes.types.prepare import PrepareConfig
    from cqpes.utils.logger import print_header

    try:
        print_header("PREPARING DATASET")

        config = PrepareConfig.from_json(config_path)
        summary = run_prepare(config, msa_path)

        summary.log()
    except Exception as e:
        print(f"\n[ERROR] Preparation failed: {e}", file=sys.stderr)
        sys.exit(1)


def _run_train(
    config_path: str,
) -> None:
    from cqpes.pipeline.train import run_train
    from cqpes.types.train import TrainConfig

    try:
        print_header("TRAINING PIP-NN")

        config = TrainConfig.from_json(config_path)
        run_train(config)

    except Exception as e:
        print(f"\n[ERROR] Training failed: {e}", file=sys.stderr)
        sys.exit(1)


def _run_export(
    workdir_path: str,
    export_type: str,
) -> None:
    from cqpes.pipeline.export import run_export

    try:
        print_header("EXPORTING MODEL")
        run_export(workdir_path, export_type)  # type: ignore
    except Exception as e:
        print(f"\n[ERROR] Export failed: {e}", file=sys.stderr)
        sys.exit(1)


def _run_test(
    workdir_path: str,
) -> None:
    from cqpes.pipeline.test import run_test
    from cqpes.utils.logger import print_header

    try:
        print_header("MODEL EVALUATION")
        run_test(workdir_path)
    except Exception as e:
        print(f"\n[ERROR] Evaluation failed: {e}", file=sys.stderr)
        sys.exit(1)


def _run_predict(
    workdir_path: str,
    xyz_path: str,
    output_path: str,
    return_au: bool,
    calc_forces: bool,
    force_mode: Literal["analytical", "numerical"],
) -> None:
    from cqpes.pipeline.predict import run_predict

    try:
        print_header("POTENTIAL PREDICTION")
        run_predict(
            workdir_path,
            xyz_path,
            output_path,
            return_au,
            calc_forces,
            force_mode,
        )
    except Exception as e:
        print(f"\n[ERROR] Prediction failed: {e}", file=sys.stderr)
        sys.exit(1)


def _run_tasks(
    args,
) -> None:
    from cqpes.pipeline.run import run_task

    try:
        run_task(args)
    except Exception as e:
        print(f"\n[ERROR] Prediction failed: {e}", file=sys.stderr)
        sys.exit(1)


if __name__ == "__main__":
    main()



import importlib.util
import sys
from types import ModuleType


def load_msa_so(
    msa_so_path: str,
) -> ModuleType:
    module_name = "msa"
    spec = importlib.util.spec_from_file_location(module_name, msa_so_path)

    if (spec is None) or (spec.loader is None):
        raise ImportError(f"Failed to create spec for msa {msa_so_path}.")

    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    return module



import os
from dataclasses import dataclass
from typing import Any, Optional, Union

import numpy as np


@dataclass
class CQPESData:
    xyz: np.ndarray  # (N, Natoms, 3)
    alpha: float
    p: np.ndarray  # (N, Npip)
    V: np.ndarray  # (N,)
    ref_energy: float

    p_min: Optional[np.ndarray] = None
    p_max: Optional[np.ndarray] = None
    V_min: Optional[float] = None
    V_max: Optional[float] = None

    def __post_init__(self) -> None:
        n_samples = self.p.shape[0]
        if self.V.shape[0] != n_samples:
            raise ValueError(
                f"V size ({self.V.shape[0]}) mismatch with "
                f"p size ({n_samples})."
            )
        if self.xyz.size > 0 and self.xyz.shape[0] != n_samples:
            raise ValueError(
                f"xyz size ({self.xyz.shape[0]}) mismatch with "
                f"p size ({n_samples})."
            )

        if self.p_min is None:
            self.p_min = self.p.min(axis=0)

        if self.p_max is None:
            self.p_max = self.p.max(axis=0)

        if self.V_min is None:
            self.V_min = self.V.min().item()

        if self.V_max is None:
            self.V_max = self.V.max().item()

    @staticmethod
    def rescale(
        t: np.ndarray,
        t_min: Union[np.ndarray, float],
        t_max: Union[np.ndarray, float],
    ) -> np.ndarray:
        with np.errstate(divide="ignore", invalid="ignore"):
            return np.nan_to_num(2 * (t - t_min) / (t_max - t_min) - 1)

    @staticmethod
    def unscale(
        scaled_t: np.ndarray,
        t_min: Union[np.ndarray, float],
        t_max: Union[np.ndarray, float],
    ) -> np.ndarray:

        return t_min + (scaled_t + 1) * (t_max - t_min) / 2

    @property
    def X(self) -> np.ndarray:
        assert (self.p_min is not None) and (self.p_max is not None)

        return self.rescale(self.p, self.p_min, self.p_max)

    @property
    def y(self) -> np.ndarray:
        assert (self.V_min is not None) and (self.V_max is not None)

        return self.rescale(self.V, self.V_min, self.V_max)

    @property
    def n_samples(self) -> int:
        return self.p.shape[0]

    @classmethod
    def from_dir(
        cls,
        path: str,
    ) -> "CQPESData":
        def _load(name: str):
            f_path = os.path.join(path, name)

            if not os.path.exists(f_path):
                raise FileNotFoundError(f"Missing: {f_path}")

            return np.load(f_path)

        return cls(
            xyz=_load("xyz.npy"),
            alpha=_load("alpha.npy").item(),
            p=_load("p.npy"),
            V=_load("V.npy"),
            ref_energy=_load("ref_energy.npy").item(),
            p_min=_load("p_min.npy"),
            p_max=_load("p_max.npy"),
            V_min=_load("V_min.npy").item(),
            V_max=_load("V_max.npy").item(),
        )

    def to_dir(
        self,
        path: str,
    ) -> str:
        if not os.path.exists(path):
            os.makedirs(path, exist_ok=True)

        payload = {
            "xyz.npy": self.xyz,
            "alpha.npy": self.alpha,
            "p.npy": self.p,
            "V.npy": self.V,
            "ref_energy.npy": self.ref_energy,
            "p_min.npy": self.p_min,
            "p_max.npy": self.p_max,
            "V_min.npy": self.V_min,
            "V_max.npy": self.V_max,
        }

        for filename, data in payload.items():
            np.save(os.path.join(path, filename), data)

        return os.path.abspath(path)

    def __getitem__(
        self,
        index: Any,
    ) -> "CQPESData":
        if isinstance(index, (int, np.integer)):
            index = [index]

        return CQPESData(
            xyz=self.xyz[index],
            alpha=self.alpha,
            p=self.p[index],
            V=self.V[index],
            ref_energy=self.ref_energy,
            p_min=self.p_min,
            p_max=self.p_max,
            V_min=self.V_min,
            V_max=self.V_max,
        )
